/* clip.c
 *	Copyright (C) 2001, 2002, 2003, 2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of MyGPGME.
 *
 * MyGPGME is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MyGPGME is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#include <windows.h>

#include "w32gpgme.h"
#include "wptCrypto.h"

/* XXX: define clipboard errors. */

/* Check if the clipboard contains text. @r_rc contains
   one text data is available.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_istext_avail (int *r_rc)
{    
    HANDLE clipmem;
    char *clipdata;
    
    if (!r_rc)
	return gpg_error (GPG_ERR_INV_ARG);
    if (OpenClipboard (NULL) == FALSE)
	return gpg_error (GPG_ERR_NO_DATA);
    
    *r_rc = 0;
    clipmem = GetClipboardData (CF_TEXT);
    if (!clipmem)
	goto leave;

    clipdata = (char *) GlobalLock (clipmem);
    if (!clipdata)
	goto leave;
    if (strlen (clipdata) > 0)
	*r_rc = 1;
    GlobalUnlock (clipmem);

leave:
    CloseClipboard ();
    return 0;
}


/* Figure out what kind of PGP data is stored in @data.
   @r_type contains the ORed types on success.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_parse_pgpid (const char *data, int *r_type)
{
    int type;

    if (!data)
	return gpg_error (GPG_ERR_INV_ARG);
    if (strlen (data) < 19) {
	*r_type = 0;
	return 0;
    }
    type = 0;
    if (strstr( data, "BEGIN PGP MESSAGE") )
	type |= PGP_MESSAGE;
    if (strstr (data, "BEGIN PGP SIGNED MESSAGE"))
	type |= PGP_CLEARSIG;
    if (strstr (data, "BEGIN PGP SIGNATURE"))
	type |= PGP_SIG;
    if (strstr (data, "BEGIN PGP PUBLIC KEY BLOCK"))
	type |= PGP_PUBKEY;
    if (strstr (data, "BEGIN PGP PRIVATE KEY BLOCK") 
	|| strstr (data, "BEGIN PGP SECRET KEY BLOCK"))
	type |= PGP_SECKEY;
    if (strstr (data, "- -----BEGIN"))
	type |= PGP_DASH_ESCAPED;
    *r_type = type;
    return 0;
}


gpgme_error_t
gpg_clip_is_secured (int *r_type, int *r_rc)
{
    int rc = 0;
    HANDLE clipmem;
    char *clipdata;
    
    if( !r_rc )
	return gpg_error (GPG_ERR_INV_ARG);
    if( OpenClipboard( NULL ) == FALSE )
	return gpg_error (GPG_ERR_INTERNAL);
    clipmem = GetClipboardData( CF_TEXT );
    if( !clipmem ) {
	*r_rc = 0;
	goto leave;
    }
    clipdata = (char *) GlobalLock( clipmem );
    if( !clipdata ) {
	*r_rc = 0;	
	goto leave;	
    }
    if( strstr( clipdata, "-----BEGIN PGP" ) 
	&& strstr( clipdata, "-----END PGP" ) )
	*r_rc = 1;
    GlobalUnlock( clipmem );
    gpg_clip_parse_pgpid( clipdata, r_type );

leave:
    CloseClipboard();
    return rc;
}


gpgme_error_t
gpg_clip_do_check( int * r_rc )
{
    int has_data = 0;
    gpgme_error_t err;

    err = gpg_clip_istext_avail (&has_data);
    if (err)
	return err;
    if (!has_data)
	return gpg_error (GPG_ERR_NO_DATA);
    gpg_clip_is_secured (NULL, r_rc);
    return 0;
}


gpgme_error_t
gpg_clip_get_pgptype (int *r_type)
{
    gpgme_error_t rc = 0;
    HANDLE clipmem;
    char *clipdata;
    
    if (OpenClipboard (NULL) == FALSE)
	return gpg_error (GPG_ERR_INTERNAL);
    clipmem = GetClipboardData (CF_TEXT);
    if (!clipmem) {
	rc = gpg_error (GPG_ERR_NO_DATA);
	goto leave;
    }
    clipdata = (char *)GlobalLock( clipmem );
    if (!clipdata) {
	rc = gpg_error (GPG_ERR_NO_DATA);
	goto leave;
    }    
    gpg_clip_parse_pgpid (clipdata, r_type);
    GlobalUnlock (clipmem);
    
leave:
    CloseClipboard ();
    return rc;
}
